/*
** Module   : ClientSocket.java
** Abstract : An interface holding the SOCKET specific methods and 
**            attributes api.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CS  20130423 Created initial version.
*/

package com.goldencode.p2j.ui.client;

import java.io.*;
import java.net.*;

/**
 * Interface for Socket methods and attributes which require to be implemented on the client
 *
 */
public interface ClientSocket
{
   /**
   * Java implementation for getting the  LOCAL-HOST SOCKET attribute which 
   * contains the address in IP form of the machine to which the Socket is
   * connected.
   * 
   * @return The address in IP form of the machine to which the Socket is
   *         connected.
   */
   public String getLocalHost();
   
   /**
    * Java implementation for getting the  LOCAL-PORT SOCKET attribute which 
    * contains the local port of the machine to which the Socket is
    * connected.
    * 
    * @return The local port of the machine to which the Socket is
    */
   public int getLocalPort();
   
   /**
    * Java implementation for getting the  LOCAL-HOST SOCKET attribute which 
    * contains the address in IP form of the machine to which the Socket is
    * connected.
    * 
    * @return The remote address in IP form of the machine to which the Socket
    *         is connected.
    */
   public String getRemoteHost();
   
   /**
    * Java implementation for getting the  REMOTE-PORT SOCKET attribute which 
    * contains the local port of the machine to which the Socket is
    * connected.
    * 
    * @return The remote port of the machine to which the Socket is connected.
    */
   public int getRemotePort();
   
   /**
    * Reads up to buffer length bytes from the socket into a byte array
    * 
    * @param  buffer
    *         The memory buffer in which the read bytes are put.
    *         
    * @return The number of bytes read
    */
   // TODO decide if this throws exception or handle it (most likely throw)
   public int read(byte[] buffer);
   
   /**
    * Writes the bute array buffer to the socket. 
    * 
    * @param  buffer
    *         The memory buffer from which the data will be written to the socket
    */
   public void write(byte[] buffer);
   
   /**
    * Method declaration  for the CONNECT() procedure supported by sockets
    * web-services and server handles.
    *  
    * @param    address
    *           The socket address to which to connect the socket.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    * @throws   IOException
    *           In case operation fails.
    */
   public boolean connect(InetSocketAddress address) throws IOException;
   
   /**
    * Method declaration  for the CONNECT() procedure supported by sockets
    * web-services and server handles.
    *  
    * @param    address
    *           The socket address to which to connect the socket.
    * @param    noSessionReuse
    *           If <code>true</code> SSL session ID will not be reused when reconnecting
    * @param    noHostVerify
    *           If <code>true</code> no host verification will be performed.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise.
    *           
    * @throws   IOException
    *           In case operation fails.
    */
   public boolean connectSsl(InetSocketAddress address,
                             boolean           noSessionReuse,
                             boolean           noHostVerify)
   throws IOException;
   
   /**
    * Sets the TCP_NODELAY socket option.
    * 
    * @param  on
    *         <code>true</code> to set enable TCP_NODELAY and <code>false<code> otherwise.
    * 
    * @return <code>true</code> if the operation succeeded, or <code>false</code>
    *         otherwise
    */
   public boolean setSocketTcpNoDelay(boolean on);
   
   /**
    * Sets the SO_LINGER socket option.
    * 
    * @param  on
    *         <code>true</code> to set enable SO_LINGER and <code>false<code> otherwise.
    * @param  linger
    *         The linger time in seconds to set.
    * 
    * @return <code>true</code> if the operation succeeded, or <code>false</code>
    *         otherwise
    */
   public boolean setSocketSoLinger(boolean on, int linger);
   
   /**
    * Sets the KEEP_ALIVE socket option.
    * 
    * @param  on
    *         <code>true</code> to set enable KEEP_ALIVE and <code>false<code> otherwise.
    * 
    * @return <code>true</code> if the operation succeeded, or <code>false</code>
    *         otherwise
    */
   public boolean setSocketKeepAlive(boolean on);
   
   /**
    * Sets the SO_REUSEADDR socket option.
    * 
    * @param  on
    *         <code>true</code> to set enable SO_REUSEADDR and <code>false<code> otherwise.
    * 
    * @return <code>true</code> if the operation succeeded, or <code>false</code>
    *         otherwise
    */
   public boolean setSocketReuseAddress(boolean on);
   
   /**
    * Sets the SO_RCVBUF socket option.
    * 
    * @param  size
    *         The receive buffer size to set.
    * 
    * @return <code>true</code> if the operation succeeded, or <code>false</code>
    *         otherwise
    */
   public boolean setSocketReceiveBufferSize(int size);
   
   /**
    * Sets the SO_SNDBUF socket option.
    * 
    * @param  size
    *         The send buffer size to set.
    * 
    * @return <code>true</code> if the operation succeeded, or <code>false</code>
    *         otherwise
    */
   public boolean setSocketSendBufferSize(int size);
   
   /**
    * Sets the socket object to use
    * 
    * @param socket
    *        The socket to set from - used from listening connections
    */
   // TODO This might not be the best solution
   public void setSocket(Socket socket);
   
   /**
    * Java implementation for getting the  SSL-SERVER-NAME SOCKET attribute 
    * which contains the SSL server name if the connection is made as SSL.
    * 
    * @return The SSL server name, if the connection is made as SSL or unknown
    *         otherwise
    */
   // TODO This needs to be checked.
   public String getSslServerName();
   
   /**
    * Java implementation for the DISCONECT() SOCKET method, this will attempt 
    * to disconnect the socket and returns <code>true</code> if succeeded or 
    * <code>false</code> otherwise.
    * 
    * @return <code>true</code> if disconnect succeeded or <code>false</code> 
    *         otherwise.
    *         
    * @throws   IOException
    *           In case operation fails.
    */
   public boolean disconnect() throws IOException;
}
