/*
** Module   : BinaryData.java
** Abstract : Progress 4GL compatible base class for raw and memptr types
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ---------------------------Description----------------------------
** 001 GES 20050906   @22562 Created initial version with support for a
**                           subset of raw and memptr processing.
** 002 ECF 20051004   @22968 Implement hashCode() method. Required to be
**                           consistent with BaseDataType's implementation
**                           of equals() method.
** 003 GES 20060206   @24296 Made compareTo() safe to process any object
**                           type (avoids ClassCastException).
** 004 GES 20060208   @25572 Added a default format string getter.
** 005 GES 20060418   @25573 Added getString() function support.
** 006 GES 20060422   @25651 Changed assignment style syntax backing
**                           methods to return "this" such that the result
**                           can be assigned back in the case of a 
**                           database field.
** 007 GES 20060427   @25768 New signatures for setString().
** 008 GES 20060724   @28157 Added a forced assign() abstract method.
** 009 GES 20060807   @28470 Moved to externalizable interface to optimize
**                           network performance.
** 010 GES 20060905   @29223 Fixed getString() to treat an over-large
**                           length value as array length minus position.
** 011 GES 20080117   @36833 Added NPE protection code.
** 012 ECF 20080312   @37440 Fixed findNextNull(). It was running off end
**                           in the event of a missing end null byte.
** 013 ECF 20080318   @37569 Fixed findNextNull() (again). It was *still*
**                           running off end when trailing null byte was
**                           missing. Thanks (and apologies) to John
**                           Ballard.
** 014 CS  20130124          Added conversion support for memptr related 
**                           functions task #1635.
** 015 GES 20130322          Modified the assign(BDT,boolean) to be the common location for _POLY 
**                           morphing. This is safe because all current usage would have
**                           generated a ClassCastException if that method was ever used with an
**                           invalid type. Now, instead of the CCE, the code will morph the value
**                           as is possible. The morphing code still needs to be checked for 
**                           validity.
** 016 CS 20130412           Added setBytes() and getByteRange(int pos, int len) methods.
*/

package com.goldencode.p2j.util;

import java.io.*;
import org.apache.commons.codec.binary.Base64;

/**
 * A class that implements common services for Progress 4GL compatible
 * raw and memptr data types whose data is mutable.  Only a subset of
 * Progress language features are supported.
 * <p>
 * In addition to the implementation of many built-in functions
 * and operators, this class hides the representation of the 
 * <code>unknown value</code> and properly maintains the processing of
 * <code>unknown value</code>. 
 * <p>
 * Format string related processing is handled in this class, see method
 * {@link #toString(String)}.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * Progress Function             Java Method
 * ---------------------------   -----------------------
 * length                        {@link #length}
 * get-byte                      {@link #getByte}
 * get-string                    {@link #getString}
 *
 * Progress Language Statement   Java Method
 * ---------------------------   -----------------------
 * length                        {@link #setLength}
 * put-byte                      {@link #setByte}
 * put-string                    {@link #setString}
 *
 * Progress Operator             Java Method
 * ---------------------------   -----------------------
 * = (assignment)                {@link #assign}
 * comparisons                   {@link #compareTo} used by {@link CompareOps} 
 * </pre> 
 * <p>
 *
 * @author    GES
 */
public abstract class BinaryData
extends BaseDataType
{
   /**
    * Stores the state of this instance as an array of bytes.  This is
    * independent of the state of whether or not this instance represents the
    * <code>unknown value</code>.
    */
   protected byte[] value = null; 
   
   /**
    * Stores the state of whether or not this instance represents the
    * <code>unknown value</code>, independent of the state of this
    * instance's data.
    */
   protected boolean unknown = false;
   
   /**
    * int variable representing the byte order on the current machine
    */
   private static final int HOST_BYTE_ORDER = 1;
   
   /** int varible representing the big endian byte order */
   private static final int BIG_ENDIAN = 2;
   
   /** int varible the little endian byte order */
   private static final int LITTLE_ENDIAN = 2;
   
   /** This represents the byte-order for the used variable*/
   private int byteOrder = HOST_BYTE_ORDER;
   
   /**
    * Default constructor, creates an instance that is uninitialized.
    */
   public BinaryData()
   {
   }
   
   /**
    * Constructor which creates an instance which is initialized based on
    * the given binary data.  The passed reference becomes the backing
    * reference for this instance (the data is not copied, only the
    * reference to the data is copied).
    *
    * @param    val
    *           The binary data to be used as the contents or if 
    *           <code>null</code> then the instance will represent the
    *           <code>unknown value</code>.
    */
   public BinaryData(byte[] val)
   {
      if (val == null)
      {
         unknown = true;
      }
      
      value = val;
   }
   
   /**
    * Constructs an instance which converts the contents of the given
    * Progress 4GL compatible string form of a raw value into the proper
    * binary form.  When Progress converts a raw value into a string,  
    * it inserts 6 hexidecimal characters at the front of the string and
    * then formats all other data as a single Base64 encoded value (it
    * is NOT split into 76 character line segments).  The first 2 hexidecimal
    * characters are always <code>02</code> and the following 4 hexidcemial
    * characters are a hexidecimal representation of the length of the
    * resulting value.  These 6 preceding bytes are simply dropped.
    *
    * @param    value
    *           The Base64 value to be converted into the resulting
    *           instance binary data.
    */
   public BinaryData(String value)
   {
      this.value = Base64.decodeBase64(value.substring(6).getBytes());
   }
   
   /**
    * Constructs an instance which converts the contents of the given
    * Progress 4GL compatible string form of a raw value into the proper
    * binary form.  When Progress converts a raw value into a string,  
    * it inserts 6 hexidecimal characters at the front of the string and
    * then formats all other data as a single Base64 encoded value (it
    * is NOT split into 76 character line segments).  The first 2 hexidecimal
    * characters are always <code>02</code> and the following 4 hexidcemial
    * characters are a hexidecimal representation of the length of the
    * resulting value.  These 6 preceding bytes are simply dropped.
    *
    * @param    value
    *           The Base64 value to be converted into the resulting
    *           instance binary data.  If this is the 
    *           <code>unknown value</code>, then the result is the
    *           <code>unknown value</code>.
    */
   public BinaryData(character value)
   {
      if (value.isUnknown())
      {
         setUnknown();
      }
      
      String str = value.getValue().substring(6);
      this.value = Base64.decodeBase64(str.getBytes());
   }
   
   /**
    * Reports if this instance represents the <code>unknown value</code>.
    *
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return unknown;
   }
   
   /**
    * Sets the state of this instance's <code>unknown value</code> flag or
    * state to <code>true</code>.
    * <p>
    * <b>Warning: the data stored in this instance may be invalid after
    * calling this method.</b>
    */
   public void setUnknown()
   {
      unknown = true;
   }   
   
   /**
    * Accesses the state of this instance's binary data independently of the
    * state of whether this instance represents the
    * <code>unknown value</code>.  If this instance represents the
    * <code>unknown value</code>, the returned <code>String</code> is
    * <b>not valid</b>! 
    *
    * @return   The state of this instance's binary data.
    */
   public byte[] getValue()
   {
      return value;
   }
   
   /**
    * Sets the state of this instance's binary data value
    * and sets the <code>unknown value</code> state for this instance to
    * <code>false</code>.
    *
    * @param    value
    *           The state to which to set this instance's binary data.
    */
   public void setValue(byte[] value)
   {
      this.value = value;
      unknown    = false;
   }
      
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * If the value is not of type <code>BinaryData</code>, the following automatic type
    * conversion will occur:
    * <p>
    * <pre>
    * TODO: implement and document this
    * </pre>
    * <p>
    * This conversion is meant to handle the cases of built-in functions and methods
    * in the 4GL which have polymorphic return types (e.g. DYNAMIC-FUNCTION()).
    *
    * @param    data
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType data, boolean force)
   {
      if (data instanceof BinaryData)
      {
         assign((BinaryData) data, force);
      }
      else
      {
         // TODO: implement type conversion here for _POLY cases 
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    data
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BinaryData data, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         // set the instance data
         this.value = duplicateBytes(data.value);
         unknown    = data.unknown;
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    data
    *           The instance from which to copy state.
    */
   public void assign(BinaryData data)
   {
      assign(data, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((BinaryData) value, false);
   }   
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   public int compareTo(Object obj)   
   {
      BinaryData bd = null;
      
      if (obj instanceof BinaryData)
      {
         bd = (BinaryData) obj;
      }
      else
      {
         return -1;
      }
      
      // handle null arrays (uninitialized memptrs)
      if (value == null)
      {
         if (bd.value == null)
         {
            return 0;
         }
         else
         {
            return -1;
         }
      }
      else
      {
         if (bd.value == null)
         {
            return 1;
         }
      }
      
      // handle 0 length arrays
      if (value.length == 0)
      {
         if (bd.value.length == 0)
         {
            return 0;
         }
         else
         {
            return -1;
         }
      }
      else
      {
         if (bd.value.length == 0)
         {
            return 1;
         }
      }
      
      // we know that both arrays are non-null and have a non-zero length      
      
      int minLen = Math.min(bd.value.length, value.length);
      
      // compare overlapping portions, first difference defines the result
      for (int i = 0; i < minLen; i++)
      {
         int result = new Byte(value[i]).compareTo(new Byte(bd.value[i])); 
         
         if (result != 0)
         {
            return result;
         }
      }
      
      // the overlapping content is equivalent, if length is the same then
      // the arrays are the same
      if (value.length == bd.value.length)
      {
         return 0;
      }
      
      // unequal lengths, the larger array is greater      
      return value.length > bd.value.length ? 1 : -1;
   }   
   
   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    * <p>
    * Note that performance is not stellar for large binary values. Likewise,
    * distribution of hash codes is fairly weak, since a maximum of 33
    * samples are used from the underlying byte array to differentiate
    * results. It is therefore recommended that large binary data values not
    * be used within or as hashing keys.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {
      int result = 17;
      if (isUnknown())
      {
         return result;
      }
      
      if (value == null || value.length == 0)
      {
         return 0;
      }
      
      int len = value.length;
      int upper = Math.min(100, len);
      for (int i = len % 3; i < upper; i += 3)
      {
         result = 37 * result + (int) value[i];
      }
      
      return result;
   }
   
   /**
    * Returns the byte at the given index position in the binary data.
    *
    * @param    pos
    *           The 1-based index position to return, must be &gt; 0.
    *
    * @return   The byte at the given index position in the binary data or
    *           the <code>unknown value</code> if this instance is
    *           <code>unknown</code>, or if the index position is
    *           <code>unknown</code> or if the index position is greater
    *           than the size of the binary data.
    *
    * @throws   ErrorConditionException
    *           If the index position is &lt; 1.
    */
   public integer getByte(NumberType pos)
   throws ErrorConditionException
   {
      if (pos.isUnknown())
      {
         return new integer();
      }
      
      return getByte(pos.intValue());
   }
      
   /**
    * Returns the byte at the given index position in the binary data.
    *
    * @param    pos
    *           The 1-based index position to return, must be &gt; 0.
    *
    * @return   The byte at the given index position in the binary data or
    *           the <code>unknown value</code> if this instance is
    *           <code>unknown</code> or if the index position is greater
    *           than the size of the binary data.
    *
    * @throws   ErrorConditionException
    *           If the index position is &lt; 1.
    */
   public integer getByte(int pos)
   throws ErrorConditionException
   {
      if (value == null)
      {
         genUninitializedError();
         return new integer();
      }
      
      if (pos < 1)
      {
         genIndexError();
         return new integer();
      }
      
      if (unknown || pos > value.length)
      {
         return new integer();
      }
      
      // convert to 0-based array index
      pos--;
      
      return new integer(value[pos] & 0x000000FF);
   }
   
   /**
    * Sets the byte at the given index position in the binary data.
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If any input is <code>unknown</code> or if
    * the instance to be modified is <code>unknown</code>, no change will
    * be made.
    *
    * @param    val
    *           The byte value to assign to the given position.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setByte(NumberType val, NumberType pos)
   throws ErrorConditionException
   {
      if (pos.isUnknown() || val.isUnknown())
      {
         return this;
      }
      
      return setByte(val.intValue(), pos.intValue());
   }
      
   /**
    * Sets the byte at the given index position in the binary data.
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If any input is <code>unknown</code> or if
    * the instance to be modified is <code>unknown</code>, no change will
    * be made.
    *
    * @param    val
    *           The byte value to assign to the given position.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setByte(NumberType val, int pos)
   throws ErrorConditionException
   {
      if (val.isUnknown())
      {
         return this;
      }
      
      return setByte(val.intValue(), pos);
   }
      
   /**
    * Sets the byte at the given index position in the binary data.
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If any input is <code>unknown</code> or if
    * the instance to be modified is <code>unknown</code>, no change will
    * be made.
    *
    * @param    val
    *           The byte value to assign to the given position.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setByte(int val, NumberType pos)
   throws ErrorConditionException
   {
      if (pos.isUnknown())
      {
         return this;
      }
      
      return setByte(val, pos.intValue());
   }
      
   /**
    * Sets the byte at the given index position in the binary data.
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code>, no change will be made.
    *
    * @param    val
    *           The byte value to assign to the given position.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setByte(int val, int pos)
   throws ErrorConditionException
   {
      if (setMethodSetup(pos, 1))
      {      
         // convert to 0-based array index
         pos--;
         
         value[pos] = (byte) (val & 0x000000FF);
      }
      
      return this;
   }
      
   /**
    * Gets the bytes at the given starting index position in the binary data
    * up to the next <code>null</code> byte or the complete contents of
    * the given buffer, whichever comes first.
    *
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0 and &lt;=
    *           the length of the buffer.
    *
    * @return   The string read or <code>unknown value</code> if any input
    *           is <code>unknown value</code> or if the index position is
    *           out of range.
    */
   public character getString(NumberType pos)
   throws ErrorConditionException
   {
      if (pos.isUnknown())
      {
         return new character();
      }
      
      return getString(pos.intValue(), -1);
   }
   
   /**
    * Gets the bytes at the given starting index position in the binary data
    * up to the next <code>null</code> byte or the complete contents of
    * the given buffer, whichever comes first.
    *
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0 and &lt;=
    *           the length of the buffer.
    *
    * @return   The string read or <code>unknown value</code> if any input
    *           is <code>unknown value</code> or if the index position is
    *           out of range.
    */
   public character getString(int pos)
   throws ErrorConditionException
   {
      return getString(pos, -1);
   }
   
   /**
    * Gets the bytes at the given starting index position in the binary data.
    * <p>
    * If the specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byte exists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    *
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0 and &lt;=
    *           the length of the buffer.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The string read or <code>unknown value</code> if any input
    *           is <code>unknown value</code> or if the index position is
    *           out of range or if the index + len is out of range.
    */
   public character getString(int pos, NumberType len)
   throws ErrorConditionException
   {
      if (len.isUnknown())
      {
         return new character();
      }
      
      return getString(pos, len.intValue());
   }
   
   /**
    * Gets the bytes at the given starting index position in the binary data.
    * <p>
    * If the specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byte exists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    *
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0 and &lt;=
    *           the length of the buffer.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The string read or <code>unknown value</code> if any input
    *           is <code>unknown value</code> or if the index position is
    *           out of range or if the index + len is out of range.
    */
   public character getString(NumberType pos, int len)
   throws ErrorConditionException
   {
      if (pos.isUnknown())
      {
         return new character();
      }
      
      return getString(pos.intValue(), len);
   }
   
   /**
    * Gets the bytes at the given starting index position in the binary data.
    * <p>
    * If the specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byte exists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    *
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0 and &lt;=
    *           the length of the buffer.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The string read or <code>unknown value</code> if any input
    *           is <code>unknown value</code> or if the index position is
    *           out of range or if the index + len is out of range.
    */
   public character getString(NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      if (pos.isUnknown() || len.isUnknown())
      {
         return new character();
      }
      
      return getString(pos.intValue(), len.intValue());
   }
   
   /**
    * Gets the bytes at the given starting index position in the binary data.
    * <p>
    * If the specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byte exists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    *
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0 and &lt;=
    *           the length of the buffer.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.  Any value that
    *           would extend past the end of the byte array is treated as if
    *           it was (size of the byte array minus the <code>pos</code>)
    *           which results in the rest of the byte array being copied into
    *           the resulting string.
    *
    * @return   The string read or <code>unknown value</code> if any input
    *           is <code>unknown value</code> or if the index position is
    *           out of range or if the index + len is out of range.
    */
   public character getString(int pos, int len)
   throws ErrorConditionException
   {
      boolean toNull = false;
      
      // special case (incorrect or unspecified length)
      if (len < 1)
      {
         // -1 is the "unspecified" length
         if (len == -1)
         {
            toNull = true;
         }
         else
         {
            // invalid index!
            genIndexError();
            return new character();
         }
      }
      
      // avoid NPE, an error will be generated in getMethodSetup
      if (value != null)
      {
         // reset the length based on the size of the array if the given
         // length is too big
         if ((pos + len) > value.length)      
         {
            // the given position is 1-based so we add 1 to account for this
            len = (value.length - pos) + 1;
         }
      }
      
      if (getMethodSetup(pos, len))
      {      
         // convert to 0-based array index
         pos--;
         
         // find ending location
         int end = toNull ? findNextNull(pos) : pos + len;
         
         // create buffer to contain the data
         byte[] text = new byte[(end - pos)];
         
         // copy the data in (may copy nothing if end == pos)
         for (int i = pos; i < end; i++)
         {
            text[i - pos] = value[i];
         }
         
         return new character(new String(text));
      }
      
      return new character();
   }
      
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  This method
    * will copy up to and including the first <code>null</code> byte or it
    * will copy to the end of the string if no <code>null</code> byte
    * exists.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(character data, NumberType pos)
   throws ErrorConditionException
   {
      if (pos.isUnknown() || data.isUnknown())
      {
         return this;
      }
      
      return setString(data.getValue(), pos.intValue(), -1);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  This method
    * will copy up to and including the first <code>null</code> byte or it
    * will copy to the end of the string if no <code>null</code> byte
    * exists.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(String data, NumberType pos)
   throws ErrorConditionException
   {
      if (pos.isUnknown())
      {
         return this;
      }
      
      return setString(data, pos.intValue(), -1);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  This method
    * will copy up to and including the first <code>null</code> byte or it
    * will copy to the end of the string if no <code>null</code> byte
    * exists.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code>, no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(String data, int pos)
   throws ErrorConditionException
   {
      return setString(data, pos, -1);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  This method
    * will copy up to and including the first <code>null</code> byte or it
    * will copy to the end of the string if no <code>null</code> byte
    * exists.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(character data, int pos)
   throws ErrorConditionException
   {
      if (data.isUnknown())
      {
         return this;
      }
      
      return setString(data.getValue(), pos, -1);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(String data, int pos, NumberType len)
   throws ErrorConditionException
   {
      if (len.isUnknown())
      {
         return this;
      }
      
      return setString(data, pos, len.intValue());
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(character data, int pos, int len)
   throws ErrorConditionException
   {
      if (data.isUnknown())
      {
         return this;
      }
      
      return setString(data.getValue(), pos, len);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(String data, NumberType pos, int len)
   throws ErrorConditionException
   {
      if (pos.isUnknown())
      {
         return this;
      }
      
      return setString(data, pos.intValue(), len);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(character data, int pos, NumberType len)
   throws ErrorConditionException
   {
      if (len.isUnknown() || data.isUnknown())
      {
         return this;
      }
      
      return setString(data.getValue(), pos, len.intValue());
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(character data, NumberType pos, int len)
   throws ErrorConditionException
   {
      if (pos.isUnknown() || data.isUnknown())
      {
         return this;
      }
      
      return setString(data.getValue(), pos.intValue(), len);
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(String data, NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      if (pos.isUnknown() || len.isUnknown())
      {
         return this;
      }
      
      return setString(data, pos.intValue(), len.intValue());
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code> or if any of the inputs are <code>unknown</code>,
    * no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(character data, NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      if (pos.isUnknown() || len.isUnknown() || data.isUnknown())
      {
         return this;
      }
      
      return setString(data.getValue(), pos.intValue(), len.intValue());
   }
   
   /**
    * Sets the bytes at the given starting index position in the binary data
    * to the partial or complete contents of the given string.  If the
    * specified <code>len</code> is -1, this method will copy up to and
    * including the first <code>null</code> byte or it will copy to the end
    * of the string if no <code>null</code> byteexists.  Otherwise, exactly
    * the specified number of characters will be copied, without regard to
    * any embedded <code>null</code> bytes.
    * <p>
    * Implements auto-extend and length limitations for subclasses which 
    * require this support.  If the instance to be modified is
    * <code>unknown</code>, no change will be made.
    *
    * @param    data
    *           Source string from which to copy.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setString(String data, int pos, int len)
   throws ErrorConditionException
   {
      boolean addNull = false;
      
      // special case (incorrect or unspecified length)
      if (len < 1)
      {
         // -1 is the "unspecified" length
         if (len == -1)
         {
            // OK, just copy the whole thing (plus add a null byte at the end)
            len = data.length() + 1;
            addNull = true;
         }
         else
         {
            // invalid index!
            genIndexError();
            return this;
         }
      }
      
      if (setMethodSetup(pos, len))
      {      
         // convert to 0-based array index
         pos--;
         
         // obtain the data as a byte[]
         StringBuffer sb = new StringBuffer();
         
         sb.append(data.substring(0, addNull ? len - 1 : len));
         
         if (addNull)
         {
            sb.append('\0');
         }
         
         byte[] text = sb.toString().getBytes();
         int    end  = pos + len;
         
         // copy the data in
         for (int i = pos; i < end; i++)
         {
            value[i] = text[i - pos];
         }
      }
      
      return this;
   }
   
   /**
    * Returns the long signed 32 bit (4 byte) value from the given 
    * position as an <code>integer<code>. 
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    * 
    * @return The BinaryDate subset from the given byte position.
    */
   public integer getLong(double pos)
   {
      return getLong(new integer(pos));
   }
   
   /**
    * Returns the long signed 32 bit (4 byte) value from the given 
    * position as an <code>integer<code>.  
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    * 
    * @return The BinaryDate subset from the given byte position.
    */
   public integer getLong(NumberType pos)
   {
      // TODO implement actual logic
      return new integer();
   }
   
   /**
    * Gets an integer as a signed 16-bit (2 byte) value in the binary data 
    * at the given position.
    *
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0.
    *    
    * @return The integer as a signed 16-bit (2 byte) value in the binary data 
    *         at the given position.
    */
   public integer getShort(NumberType pos)
   {
      return getShort(pos.intValue());
   }
   
   /**
    * Gets an integer as a signed 16-bit (2 byte) value in the binary data 
    * at the given position.
    *
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0.
    *    
    * @return The integer as a signed 16-bit (2 byte) value in the binary data 
    *         at the given position.
    */
   public integer getShort(double pos)
   {
      // TODO implement actual logic
      return new integer();
   }
   
   /**
    * Puts an integer as a signed 16-bit (2 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 16-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setShort(double data, double pos)
   {
      setShort(new integer(data), new integer(pos));
   }
   
   /**
    * Puts an integer as a signed 16-bit (2 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 16-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setShort(NumberType data, double pos)
   {
      setShort(data,new integer(pos));
   }
   
   /**
    * Puts an integer as a signed 16-bit (2 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 16-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setShort(double data, NumberType pos)
   {
      setShort(new integer(data), pos);
   }
   /**
    * Puts an integer as a signed 16-bit (2 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 16-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setShort(NumberType data, NumberType pos)
   {
      // TODO implement actual logic
   }
   
   /**
    * Puts an integer as a signed 32-bit (4 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 32-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setLong(double data, double pos)
   {
      setLong(new integer(data), new integer(data));
   }
   
   /**
    * Puts an integer as a signed 32-bit (4 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 32-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setLong(NumberType data, double pos)
   {
      setLong(data, new integer(pos));
   }
   
   /**
    * Puts an integer as a signed 32-bit (4 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 32-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setLong(double data, NumberType pos)
   {
      setLong(new integer(data), pos);
   }
   
   /**
    * Puts an integer as a signed 32-bit (4 byte) value in the binary data 
    * from the given position.
    *
    * @param data
    *        The 32-bit <code>integer</code> value to be set
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setLong(NumberType data, NumberType pos)
   {
      // TODO implement actual logic;
   }

   /**
    * Sets the binary data to point to a memory address
   
    * @param pointer
    *        The <code>integer</code> 32 bit value representing the pointer 
    *        to the memory address.
    */
   public void setPointerValue(NumberType pointer)
   {
      // TODO Implement the actual logic
   }
   
   /**
    * Set the memptr to point to a memory address
   
    * @param pointer
    *        The <code>integer</code> 32 bit value representing the pointer 
    *        to the memory address.
    */
   public void setPointerValue(double pointer)
   {
      setPointerValue(new integer(pointer));
   }
   
   /**
    * Method implementation of the GET-POINTER-VALUE 4GL instruction which
    * returns a 32 bit integer value representing the 4 byte pointer to the 
    * memptr variable.
    * 
    * @return A 32 bit integer value representing the 4 byte pointer to the 
    * memptr variable.
    */
   public integer getPointerValue()
   {
      // TODO Implement the actual logic
      return new integer();
   }
   
   /**
    * Sets a variable indicating the byte order pointed to by the memptr 
    * variable.
    * <p>
    * The possible byte orders are specified by the static variables:
    * HOST_BYTE_ORDER, BIG_ENDIAN or LITTLE_ENDIAN.
    * 
    * @param order
    *        The <code>int<code> variable used to set the byte-order mode
    */
   public void setByteOrder(int order)
   {
      // TODO Implement the actual logic
      // if the value is not HOST_BYTE_ORDER BIG_ENDIAN or LITTLE_ENDIAN 
      // error Invalid value for SET-BYTE-ORDER (9259) will be thrown
      byteOrder = order;
   }
   /**
    * Sets a variable indicating the byte order in which the memptr variable
    * is used.
    * <p>
    * The possible byte orders are specified by the static variables:
    * HOST_BYTE_ORDER, BIG_ENDIAN or LITTLE_ENDIAN.
    * 
    * @param order
    *        The <code>integer<code> variable used to set the byte-order mode
    */
   public void setByteOrder(NumberType order)
   {
      // TODO Implement the actual logic
      // if the value is not HOST_BYTE_ORDER BIG_ENDIAN or LITTLE_ENDIAN 
      // error Invalid value for SET-BYTE-ORDER (9259) will be thrown
      if (order == null || order.isUnknown())
      {
         // unknown case, usually this will set the instance to be unknown, but it must
         // simply do whatever the 4GL behavior is
      }
      else
      {
         byteOrder = order.intValue();
      }
   }

   /**
    * Returns the index representing the byte order in which this memptr 
    * variable can be used.
    * <p>
    * If this variable is not set, the default value will be returned, 
    * specified by HOST_BYTE_ORDER static variable.
    * 
    * @return The integer value which indicates the byte-order setting of the
    *         memptr variabl.
    */
   public integer getByteOrder()
   {
      // TODO Implement the actual logic
      return new integer(byteOrder);
   }

   /**
    * GET-DOUBLE function implementation which returns the decimal value 
    * representing the 8 byte floating point at the specified binary data 
    * location.
    * 
    * @param  pos
    *         An <code>integer</code> value representing the position from 
    *         where the information will be taken.
    *         
    * @return The decimal value representing the 8 byte floating point at the 
    *         specified binary data location
    */
   public decimal getDouble(NumberType pos)
   {
      // TODO Implement the actual logic
      return new decimal();
   }

   /**
    * GET-DOUBLE function implementation which returns the decimal value 
    * representing the 8 byte floating point at the specified binary data 
    * location.
    * 
    * @param  pos
    *         An <code>int</code> value representing the position from 
    *         where the information will be taken.
    *         
    * @return The decimal value representing the 8 byte floating point at the 
    *         specified binary data location
    */
   public decimal getDouble(double pos)
   {
      return getDouble(new integer(pos));
   }

   
   /**
    * Sets the decimal representing the 8 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 8 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setDouble(double data, double pos)
   {
      setDouble(new decimal(data), new integer(pos));
      
   }
   /**
    * Sets the decimal representing the 8 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 8 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setDouble(double data, NumberType pos)
   {
      setDouble(new decimal(data), pos);
   }
   
   /**
    * Sets the decimal representing the 8 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 8 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setDouble(NumberType data, double pos)
   {
      setDouble(data, new integer(pos));
   }
   
   /**
    * Sets the decimal representing the 8 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 8 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setDouble(NumberType data, NumberType pos)
   {
      // TODO Implement the actual logic
   }
  
   /**
    * Sets the decimal representing the 4 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 4 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setFloat(double data, double pos)
   {
      setFloat(new decimal(data), new integer(pos));
   }
   
   /**
    * Sets the decimal representing the 4 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 4 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setFloat(double data, NumberType pos)
   {
      setFloat(new decimal(data), pos);
   }
   

   
   /**
    * Sets the decimal representing the 4 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 4 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setFloat(NumberType data, double pos)
   {
      setFloat(data, new integer(pos));
   }
   
   /**
    * Sets the decimal representing the 4 byte floating-point value 
    * at the specified binary data location.
    * 
    * @param data
    *        The 4 byte floating point value to be set at the specified 
    *        binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setFloat(NumberType data, NumberType pos)
   {
      // TODO Implement the actual logic
   }
   
   /**
    * GET-FLOAT function implementation which returns the decimal value 
    * representing the 4 byte floating point at the specified binary data 
    * location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The decimal value representing the 4 byte floating point at the 
    *         specified binary data location
    */
   public decimal getFloat(NumberType pos)
   {
      // TODO Implement the actual logic
      return new decimal();
   }
   
   /**
    * GET-FLOAT function implementation which returns the decimal value 
    * representing the 4 byte floating point at the specified binary data 
    * location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The decimal value representing the 4 byte floating point at the 
    *         specified binary data location
    */
   public decimal getFloat(double pos)
   {
      // TODO Implement the actual logic
      return getFloat(new integer(pos));
   }
   
   /**
    * Sets an int64 or integer as a signed 64 bit value at the specified 
    * binary data location
    * 
    * @param data
    *        An int64 or integer to be set as a 64 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setInt64(double data, double pos)
   {
      setInt64(new int64(data) , new integer(pos));
   }
   
   /**
    * Sets an int64 or integer as a signed 64 bit value at the specified 
    * binary data location
    * 
    * @param data
    *        An int64 or integer to be set as a 64 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setInt64(double data, NumberType pos)
   {
      setInt64(new int64(data) , pos);
   }
   
   /**
    * Sets an int64 or integer as a signed 64 bit value at the specified 
    * binary data location
    * 
    * @param data
    *        An int64 or integer to be set as a 64 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setInt64(integer data, double pos)
   {
      setInt64(new int64(data) , pos);
   }
   
   /**
    * Sets an int64 or integer as a signed 64 bit value at the specified 
    * binary data location
    * 
    * @param data
    *        An int64 or integer to be set as a 64 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setInt64(NumberType data, NumberType pos)
   {
      // TODO Implement the actual logic
   }
   
   /**
    * Sets an int64 or integer as a signed 64 bit value at the specified 
    * binary data location
    * 
    * @param data
    *        An int64 or integer to be set as a 64 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setInt64(NumberType data, double pos)
   {
      setInt64(data , new integer(pos));
   }
   
   /**
    * Gets the int64 as a signed 64 bit value from the specified binary data
    * location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The int64 as a signed 64 bit value from the specified binary 
    *         data location.
    */
   public int64 getInt64(NumberType pos)
   {
      // TODO Implement the actual logic
      return new int64();
   }
   
   /**
    * Gets the int64 as a signed 64 bit value from the specified binary data
    * location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The int64 as a signed 64 bit value from the specified binary 
    *         data location.
    */
   public int64 getInt64(double pos)
   {
      return getInt64(new integer(pos));
   }

   /**
    * Sets an integer as an unsigned 16 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An integer to be set as a 16 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedShort(double data, double pos)
   {
      setUnsignedShort(new integer(data), new integer(pos));
   }
   
   /**
    * Sets an integer as an unsigned 16 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An integer to be set as a 16 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedShort(double data, NumberType pos)
   {
      setUnsignedShort(new integer(data), pos);
      
   }
   
   /**
    * Sets an integer as an unsigned 16 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An integer to be set as a 16 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedShort(NumberType data, double pos)
   {
      setUnsignedShort(data , new integer(pos));
   }
   
   /**
    * Sets an integer as an unsigned 16 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An integer to be set as a 16 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedShort(NumberType data, NumberType pos)
   {
      // TODO Implement the actual logic
   }
   
   /**
    * Gets the integer as an unsigned 16 bit value from the specified binary 
    * data location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The integer as a unsigned 16 bit value from the specified binary 
    *         data location.
    */
   public integer getUnsignedShort(NumberType pos)
   {
      // TODO Implement the actual logic
      return new integer();
   }
   
   /**
    * Gets the integer as an unsigned 16 bit value from the specified binary 
    * data location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The integer as a unsigned 16 bit value from the specified binary 
    *         data location.
    */
   public integer getUnsignedShort(double pos)
   {
      return getUnsignedShort(new integer(pos));
   }
   
   /**
    * Sets an int64 or integer as an unsigned 32 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An int64 or integer to be set as a 32 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedLong(double data, double pos)
   {
      setUnsignedLong(new int64(data), new integer(pos));
   }
   
   /**
    * Sets an int64 or integer as an unsigned 32 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An int64 or integer to be set as a 32 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedLong(double data, NumberType pos)
   {
      setUnsignedLong(new int64(data), pos);
   }
   
   /**
    * Sets an int64 or integer as an unsigned 32 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An int64 or integer to be set as a 32 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedLong(NumberType data, double pos)
   {
      setUnsignedLong(new int64(data), new integer(pos));
   }
   
   /**
    * Sets an int64 or integer as an unsigned 32 bit value at the specified 
    * binary data location.
    * 
    * @param data
    *        An int64 or integer to be set as a 32 bit value at the 
    *        specified  binary data location. 
    * @param pos
    *        The 1-based index position to set, must be &gt; 0.
    */
   public void setUnsignedLong(NumberType data, NumberType pos)
   {
      // TODO Implement the actual logic
   }
   
   /**
    * Gets the int64 as an unsigned 32 bit value from the specified binary 
    * data location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The int64 as a signed 64 bit value from the specified binary 
    *         data location.
    */
   public int64 getUnsignedLong(NumberType pos)
   {
      return getUnsignedLong(pos);
   }
   
   /**
    * Gets the int64 as an unsigned 32 bit value from the specified binary 
    * data location.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer..
    *         
    * @return The int64 as a signed 64 bit value from the specified binary 
    *         data location.
    */
   public int64 getUnsignedLong(double pos)
   {
      return getUnsignedLong(new integer(pos));
   }
 
   /**
    * Copies the binary data to the specified location of the current binary
    * data.
    *
    * @param    data
    *           The binary data to be copied.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    */
   public void setBytes(BinaryData data, double pos, double len)
   {
      setBytes(data, new integer(pos), new integer(len));
   }
   
   /**
    * Copies the binary data to the specified location of the current binary
    * data.
    *
    * @param    data
    *           The binary data to be copied.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    */
   public void setBytes(BinaryData data, NumberType pos, double len)
   {
      setBytes(data, pos, new integer(len));
   }
   
   /**
    * Copies the binary data to the specified location of the current binary
    * data.
    *
    * @param    data
    *           The binary data to be copied.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    */
   public void setBytes(BinaryData data, double pos, NumberType len)
   {
      setBytes(data, new integer(pos), len);
   }
   
   /**
    * Copies the binary data to the specified location of the current binary
    * data.
    *
    * @param    data
    *           The binary data to be copied.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    */
   public void setBytes(BinaryData data, NumberType pos, NumberType len)
   {
      // TODO unknown parameter error checking
      // TODO improve it?
      if (data.isUnknown() || pos.isUnknown() || len.isUnknown())
      {
         //TODO Error here
         return;
      }
      setByteRange(data.value, pos.intValue(),len.intValue());

   }
   
   /**
    * Sets the byte array data from the specified location in the current binary
    * data.
    *
    * @param    source
    *           The source byte array data to be set.
    * @param    pos
    *           The 1-based index position to set, must be &gt; 0.
    * @param    len
    *           The number of bytes to copy or -1 to copy up to and
    *           including the first <code>null</code> byte.
    */
   public void setByteRange(byte[] source, int pos, int len)
   {
      int javaPos = pos - 1;
      if (javaPos < 0 || len < 0)
      {
          // TODO error here;
         return;
      }
      
      int realPos = javaPos + len > value.length ? value.length : javaPos + len;
      int realLength = (javaPos + len) < source.length ? len : source.length - javaPos;
      
      realLength = realLength < source.length ? realLength : source.length;
      int j = 0;
      
      for (int i = javaPos; i < realPos; i++)
      {
         value[i] = source[j];
         j++;
      }
   }

   /**
    * Gets the bytes at the given index position from the binary data.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer.
    * @param  len
    *         The number of bytes to copy or -1 to copy up to and
    *         including the first <code>null</code> byte.  
    *         
    * @return The binary data with the specified length found at the given 
    *         position.
    */
   public BinaryData getBytes(NumberType pos, NumberType len)
   {
      // TODO Implement the actual logic
      return null;
   }
   
   /**
    * Gets the bytes at the given index position from the binary data.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer.
    * @param  len
    *         The number of bytes to copy or -1 to copy up to and
    *         including the first <code>null</code> byte.  
    *         
    * @return The binary data with the specified length found at the given 
    *         position.
    */
   public BinaryData getBytes(double pos, NumberType len)
   {
      return getBytes(new integer(pos), len);
   }
   
   /**
    * Gets the bytes at the given index position from the binary data.
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer.
    * @param  len
    *         The number of bytes to copy or -1 to copy up to and
    *         including the first <code>null</code> byte.  
    *         
    * @return The binary data with the specified length found at the given 
    *         position.
    */
   public BinaryData getBytes(NumberType pos, double len)
   {
      return getBytes(pos, new integer(len));
   }
   
   /**
    * Gets the bytes at the given index position from the binary data
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer.
    * @param  len
    *         The number of bytes to copy or -1 to copy up to and
    *         including the first <code>null</code> byte.  
    *         
    * @return The binary data with the specified length found at the given 
    *         position.
    */
   public BinaryData getBytes(double pos, double len)
   {
      return getBytes(new integer(pos), new integer(len));
   }
   
   /**
    * Returns a byte array copy of this variable at the given position from this variable
    * 
    * @param  pos
    *         The 1-based index position to set, must be &gt; 0 and &lt;=
    *         the length of the buffer.
    * @param  len
    *         The number of bytes to copy or -1 to copy up to and
    *         including the first <code>null</code> byte. 
    *          
    * @return The byte array copy of this variable at the given position from this variable.
    */
   public byte[] getByteRange(int pos, int len)
   {
      int javaPos = pos -1; 
      if (javaPos < 0 || len < 0)
      {
          // TODO error here;
         return null;
      }
      int realPos = javaPos + len > value.length ? value.length : javaPos + len;
      int realLen = javaPos + len > value.length ? value.length - javaPos : len;
     
      byte b[] = new byte[realLen];
      int j = 0;
      
      for (int i = javaPos; i < realPos; i++)
      {
         b[j] = value[i];
         j++;
      }
      
      return b;
   }
      
   /**
    * Returns the length of the given instance in bytes.
    *
    * @param    bd
    *           The instance to test.
    *
    * @return   The number of bytes in the given instance or
    *           <code>unknown</code>.
    */
   public static integer length(BinaryData bd)
   {
      if (bd.unknown)
         return new integer();
      
      // can't be used with memptr so a null check is not needed
      return new integer(bd.value.length);
   }
   
   /**
    * Returns the length this instance's data in bytes.
    *
    * @return   The number of bytes in this instance or <code>unknown</code>.
    */
   public integer length()
   {
      if (unknown)
         return new integer();
      
      // can't be used with memptr so a null check is not needed
      return new integer(value.length);
   }
   
   /**
    * Returns the length this instance's data in bytes.
    *
    * @return   The number of bytes in this instance or -1 if this instance
    *           is <code>unknown</code>.
    */
   public int lengthOf()
   {
      if (unknown)
         return -1;
      
      // can't be used with memptr so a null check is not needed
      return value.length;
   }
   
   /**
    * Sets the array to a given length, which may result in either extending
    * or truncating the array.
    *
    * @param    len
    *           New length of the array.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setLength(int len)
   {
      // can't be used with memptr so a null check is not needed
      value = extendBytes(value, len);
      return this;
   }
   
   /**
    * Sets the array to a given length, which may result in either extending
    * or truncating the array.
    *
    * @param    len
    *           New length of the array.
    *
    * @return   The current instance is returned.
    */
   public BinaryData setLength(NumberType len)
   {
      if (!len.isUnknown())
      {
         setLength(len.intValue());
      }
      
      return this;
   }
   
   /**
    * Return the default display format string for this type.
    * <p>
    * Warning: this always returns the empty string as this type is not
    * directly displayable.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()   
   {
      return "";
   }   
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be limited in length.
    *
    * @return    <code>true</code> if the length must be limited.
    */
   public abstract boolean isLengthLimited();
   
   /**
    * Determines the maximum length of the array of binary data stored in an
    * instance of this class.
    *
    * @return    The length limit for binary data or -1 if this class has
    *            no limit.
    */
   public abstract int lengthLimit();
   
   /**
    * Determines if the array of binary data stored in an instance of this
    * class must be automatically extended in length when operations
    * write past the current length.  Such extensions will pad with
    * <code>null</code> bytes.
    *
    * @return   <code>true</code> if the length must automatically extended.
    */
   public abstract boolean isAutoExtend();
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      unknown = in.readBoolean();
      
      // the next object can be null so we read an indicator var first to
      // determine if there is following data to be read
      
      if (in.readBoolean())
         value = (byte[]) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeBoolean(unknown);
      
      // the next object can be null so we write an indicator var first to
      // notify reader if there is following data to be read
      
      boolean there = (value != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(value);
   }
   
   /**
    * Helper to create a new byte array with the same contents as the given
    * byte array.  This is safe to use on 0 length arrays.
    *
    * @param    src
    *           The array to copy.
    *
    * @return   The resulting copy of the input array or <code>null</code>
    *           if the input is <code>null</code>.
    */
   protected static byte[] duplicateBytes(byte[] src)
   {
      if (src == null)
         return null;
      
      byte[] array = new byte[src.length];
      
      // copy the array
      for (int i = 0; i < src.length; i++)
      {
         array[i] = src[i];
      }
      
      return array;
   }
   
   /**
    * Helper to create a new byte array with the same contents as the given
    * byte array and the specified length.  If the given target length
    * is smaller than the source array length, the output array is truncated    
    * to the specified size.  If the target length is larger than the
    * source array length, the output array is extended to the given size
    * and any additional bytes are initialized to '\0'.
    * <p>
    * This is safe to use on 0 length arrays.
    *
    * @param    src
    *           The array to copy (should not be <code>null</code>).
    * @param    len
    *           The output array length.
    *
    * @return   The resulting copy of the input array. 
    */
   protected static byte[] extendBytes(byte[] src, int len)
   {
      byte[] array = new byte[len];
      
      // copy the array and also honor a shorter length if specified
      for (int i = 0; i < src.length && i < len; i++)
      {
         array[i] = src[i];
      }
      
      // if we are extending, pad with nulls (we drop through if there is
      // nothing to do)
      for (int i = src.length; i < len; i++)
      {
         array[i] = (byte) 0;
      }
      
      return array;
   }
   
   /**
    * Common worker to handle the error checking/processing with no triggering
    * of auto-expand features.
    *
    * @param    first
    *           The 1-based index position of first access, must be &gt; 0.
    * @param    len
    *           The number of bytes to be accessed starting at 
    *           <code>pos</code>, must be &gt; 0.
    *
    * @return   <code>true</code> if no errors occurred and the calling
    *           set method should continue. <code>false</code> if the
    *           calling method should immediately and silently return.
    *
    * @throws   ErrorConditionException
    *           When any of the tested failure conditions are encountered.
    */
   private boolean getMethodSetup(int first, int len)
   throws ErrorConditionException
   {
      if (unknown)
      {
         return false;
      }
      
      if (value == null)
      {
         genUninitializedError();
         return false;
      }
      
      if (first < 1)
      {
         genIndexError();
         return false;
      }
      
      // index of last byte to be accessed
      int last = (first + len - 1);
      
      if (first > value.length) 
      {
         genOutOfBoundsError();
         return false;
      }
      
      if (last > value.length)
      {
         return false;
      }
      
      return true;   
   }
   
   /**
    * Returns the next index position into the array which is the 
    * <code>null</code> byte, starting at a given index position.
    *
    * @param    pos
    *           The index into the array at which to start searching.
    *
    * @return   The next index into the array which is the <code>null</code> 
    *           byte or the length of the array if no <code>null</code> byte
    *           exists.
    */
   private int findNextNull(int pos)
   {
      int i = pos;
      
      // not all instances will end in null byte, so check length first to
      // prevent overrun
      while (i < value.length && value[i] != '\0')
      {
         i++;
      }
      
      return i;
   }
   
   /**
    * Common worker to handle the error checking/processing and the honoring
    * of length/auto-expand features.
    *
    * @param    first
    *           The 1-based index position of first access, must be &gt; 0.
    * @param    len
    *           The number of bytes to be accessed starting at 
    *           <code>pos</code>, must be &gt; 0.
    *
    * @return   <code>true</code> if no errors occurred and the calling
    *           set method should continue. <code>false</code> if the
    *           calling method should immediately and silently return.
    *
    * @throws   ErrorConditionException
    *           When any of the tested failure conditions are encountered.
    */
   private boolean setMethodSetup(int first, int len)
   throws ErrorConditionException
   {
      if (unknown)
      {
         return false;
      }
      
      if (value == null)
      {
         genUninitializedError();
         return false;
      }
      
      if (first < 1)
      {
         genIndexError();
         return false;
      }
      
      // index of last byte to be accessed
      int last = (first + len - 1);
      
      // special fixup mode is possible here
      if (first > value.length || last > value.length) 
      {
         if (isAutoExtend())
         {
            // error if this is too long
            if (isLengthLimited())
            {
               int limit = lengthLimit();
               
               if (first > limit || last > limit) 
               {
                  genLimitedLengthError();
                  return false;
               }
            }
            
            // extend and pad with nulls
            value = extendBytes(value, last);
         }
         else
         {            
            // should be only for MEMPTR
            genOutOfBoundsError();
            return false;
         }
      }
      
      return true;   
   }
   
   /**
    * Generate the standard error message in the case where an invalid
    * index position (&lt; 1) is used.
    */
   private void genIndexError()
   throws ErrorConditionException
   {
      String err = "Byte position for PUT-BYTE, GET-BYTE and other" +
                   " raw functions must be 1 or greater";
      
      ErrorManager.recordOrThrowError(1830, err);
   }
   
   /**
    * Generate the standard error message in the case where an access is
    * requested for an uninitialized variable.
    */
   private void genUninitializedError()
   throws ErrorConditionException
   {
      String err = "Put/Get functions not allowed on uninitialized memory";
      ErrorManager.recordOrThrowError(2914, err);
   }
   
   /**
    * Generate the standard error message in the case where an access is
    * requested for a location past the end of an initialized
    * <code>RAW</code> variable.
    */
   private void genLimitedLengthError()
   throws ErrorConditionException
   {
      String err = "PUTBYTE/LENGTH assignment too long -- increase -s" +
                   " parameter";
      ErrorManager.recordOrThrowError(1829, err);
   }
   
   /**
    * Generate the standard error message in the case where an access is
    * requested for a location past the end of an initialized
    * <code>MEMPTR</code> variable.
    */
   private void genOutOfBoundsError()
   throws ErrorConditionException
   {
      String err = "Can't PUT past the end of the MEMPTR";
      ErrorManager.recordOrThrowError(4791, err);
   }
}
